import java.util.*;

/* WalleBoard

  11 lignes par 9 colonnes.
  ligne 0 au nord, colonne 0 à l'ouest

  Les cases de la lignes 0 servent de position initiale
  au joueur 1, et celle de la ligne 10 pour le joueur 2.

  Chaque case est une byte qui vaut 0 si aucun mur.
  et 15 si entourée de murs, avec :
  mur est = 1,
  mur nord = 2,
  mur ouest = 4
  mur sud = 8

  Le plateau est entouré par un mur.
  
 */
class WalleBoard extends GameState {

	private int[][] board;
	private int[] lig; //lig[0] = ligne du joueur 1, lig[1] = ligne du joueur 2
	private int[] col; //col[0] = colonne du joueur 1, col[1] = colonne du joueur 2
	private int[] nbWalls; // number of remaining walls for each player

	final static int WALL_E = 1;
	final static int WALL_N = 2;
	final static int WALL_W = 4;
	final static int WALL_S = 8;

	public WalleBoard() {

		nbWalls = new int[2];
		nbWalls[0] = 10;
		nbWalls[1] = 10;

		board = new int[11][9]; // 1 dim = ligne, 2 dim = colonne
		for(int i=0;i<11;i++) {
			for(int j=0;j<9;j++) {
				if (j == 0) {
					board[i][j] = WALL_W;
				}
				else if (j == 8) {
					board[i][j] = WALL_E;
				}
				else {
					board[i][j] = 0;
				}
				if (i == 0) {
					board[0][j] |= WALL_N;
				}
				if (i == 10) {
					board[10][j] |= WALL_S;
				}
			}
		}
		lig = new int[2];
		lig[0] = -1;
		lig[1] = -1;
		col = new int[2];
		col[0] = -1;
		col[1] = -1;
	}

	public void print() {

		for (int i = 0; i < 11; i++) {
			for (int j = 0; j < 9; j++) {
				System.out.print((board[i][j] & 255)+" ");
			}
			System.out.println();
		}
	}

	public GameState clone() {
		WalleBoard newOne = new WalleBoard();
		for(int i=0;i<11;i++) {
			for(int j=0;j<9;j++) {
				newOne.board[i][j] = board[i][j];
			}
		}
		newOne.lig[0] = lig[0];
		newOne.lig[1] = lig[1];
		newOne.col[0] = col[0];
		newOne.col[1] = col[1];
		newOne.nbWalls[0] = nbWalls[0];
		newOne.nbWalls[1] = nbWalls[1];
		return newOne;
	}

	public int[][] getBoardState() {
		int[][] b = new int[11][9];
		for(int i=0;i<11;i++) {
			for(int j=0;j<9;j++) {
				b[i][j] = board[i][j];
			}
		}
		return b;
	}

	/* currentPosition() :
		- returns the position of idPlayer as an int array
		NB : if I know my id, then the id of the other
		player is (id+1)%2. Thus, it is possible for a player to retrieve
		the position of the other player.
	 */
	public int[] currentPosition(int idPlayer) {
		int[] pos = new int[2];
		pos[0] = lig[idPlayer];
		pos[1] = col[idPlayer];
		return pos;
	}

	public int remainingWalls(int idPlayer) {
		return nbWalls[idPlayer];
	}

	public boolean isNorthWall(int l, int c) {
		return (board[l][c] & WALL_N)==0?false:true;
	}
	public boolean isSouthWall(int l, int c) {
		return (board[l][c] & WALL_S)==0?false:true;
	}
	public boolean isEastWall(int l, int c) {
		return (board[l][c] & WALL_E)==0?false:true;
	}
	public boolean isWestWall(int l, int c) {
		return (board[l][c] & WALL_W)==0?false:true;
	}
	/* initPosition(): allows to set the initial position of a player's pawn
		- return 0 it it is possible, -1 else.

	 */
	public int initPosition(int idPlayer, int l, int c) {
		if ((idPlayer == 0) && (l != 0)) {
			return -1;
		}
		else if ((idPlayer == 1) && (l != 10)) {
			return -1;
		}
		else if ((c<0) || (c>8)) {
			return -1;
		}
		lig[idPlayer] = l;
		col[idPlayer] = c;
		return 0;
	}

	/* NOTES:

       Un pion ne se déplace que d'une case vers le nord, sud, est, ouest
       à condition qu'il n'y ait pas de mur s'opposant au déplacement.

       Exception : le pion de l'autre joueur se trouve sur une case voisine.
       Dans ce cas :
          - pour la direction de dplt choisie, il n'y a pas de mur "derrière" le pion
      de l'autre joueur -> le pion du joueur courant fait 2 case dans cette direction.
      - s'il y a un mur, le pion du joueur courant peut aller en oblique, ce qui laisse
      2 choix possibles, sauf s'il y a de nouveau un mur qui s'oppose.
       
     */
	public int movePawn(int idPlayer, int l, int c) {

		int ret = canMovePawn(idPlayer,l,c);
		if (ret == 0) {
			lig[idPlayer] = l;
			col[idPlayer] = c;
			if ((idPlayer == 0) && (l == 10)) {
				return 1;
			}
			else if ((idPlayer == 1) && (l == 0)) {
				return 2;
			}
			return 0;
		}
		return -1;
	}

	public int canMovePawn(int idPlayer, int l, int c) {

		int idOther = (idPlayer+1)%2;

		// test si case à atteindre en dehors du plateau
		if ((l<0) || (c<0) || (l>10) || (c>8)) {
			return -1;
		}


		// test si l,c = coord joueur => erreur
		if ((l == lig[idPlayer]) && (c == col[idPlayer])) {
			return -1;
		}
		// test si l,c = coord autre joueur => erreur
		if ((l == lig[idOther]) && (c == col[idOther])) {
			return -1;
		}

		// saut vers le sud
		if ((c == col[idPlayer]) && (l == lig[idPlayer]+2)) {
			// si pas de murs au sud
			if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_S) == 0) &&
					((board[lig[idPlayer]+1][col[idPlayer]] & WALL_S) == 0) ) {
				// si autre joueur au sud
				if ((col[0] == col[1]) && (lig[idOther] == lig[idPlayer]+1)) {
					return 0;
				}
			}
		}
		// saut vers le nord
		else if ((c == col[idPlayer]) && (l == lig[idPlayer]-2)) {
			// si pas de murs au nord
			if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_N) == 0) &&
					((board[lig[idPlayer]-1][col[idPlayer]] & WALL_N) == 0) ) {
				// si autre joueur au nord
				if ((col[0] == col[1]) && (lig[idOther] == lig[idPlayer]-1)) {
					return 0;
				}
			}
		}
		// saut vers l'est
		else if ((c == col[idPlayer]+2) && (l == lig[idPlayer])) {
			// si pas de murs à l'est
			if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_E) == 0) &&
					((board[lig[idPlayer]][col[idPlayer]+1] & WALL_E) == 0) ) {
				// si autre joueur à l'est
				if ((lig[0] == lig[1]) && (col[idOther] == col[idPlayer]+1)) {
					return 0;
				}
			}
		}
		// saut vers l'ouest
		else if ((c == col[idPlayer]-2) && (l == lig[idPlayer])) {
			// si pas de murs à l'ouest
			if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_W) == 0) &&
					((board[lig[idPlayer]][col[idPlayer]-1] & WALL_W) == 0) ) {
				// si autre joueur à l'ouest
				if ((lig[0] == lig[1]) && (col[idOther] == col[idPlayer]-1)) {
					return 0;
				}
			}
		}

		// saut oblique sud-est
		if ((c == col[idPlayer]+1) && (l == lig[idPlayer]+1)) {
			// si autre joueur à l'est
			if ((lig[0] == lig[1]) && (col[idOther] == col[idPlayer]+1)) {
				// si mur à l'est de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_E) != 0) {
					// si pas de mur à l'est puis sud
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_E) == 0) &&
							((board[lig[idPlayer]][col[idPlayer]+1] & WALL_S) == 0) ) {
						return 0;
					}
				}
			}
			// si autre joueur au sud
			else if ((col[0] == col[1]) && (lig[idOther] == lig[idPlayer]+1)) {
				// si mur au sud de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_S) != 0) {
					// si pas de mur au sud puis à l'est
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_S) == 0) &&
							((board[lig[idPlayer]+1][col[idPlayer]] & WALL_E) == 0) ) {
						return 0;
					}
				}
			}
		}
		// saut oblique nord-est
		else if ((c == col[idPlayer]+1) && (l == lig[idPlayer]-1)) {
			// si autre joueur à l'est
			if ((lig[0] == lig[1]) && (col[idOther] == col[idPlayer]+1)) {
				// si mur à l'est de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_E) != 0) {
					// si pas de mur à l'est puis nord
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_E) == 0) &&
							((board[lig[idPlayer]][col[idPlayer]+1] & WALL_N) == 0) ) {
						return 0;
					}
				}
			}
			// si autre joueur au nord
			else if ((col[0] == col[1]) && (lig[idOther] == lig[idPlayer]-1)) {
				// si mur au nord de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_N) != 0) {
					// si pas de mur au nord puis à l'est
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_N) == 0) &&
							((board[lig[idPlayer]-1][col[idPlayer]] & WALL_E) == 0) ) {
						return 0;
					}
				}
			}
		}
		// saut oblique sud-ouest
		else if ((c == col[idPlayer]-1) && (l == lig[idPlayer]+1)) {
			// si autre joueur à l'ouest
			if ((lig[0] == lig[1]) && (col[idOther] == col[idPlayer]-1)) {
				// si mur à l'ouest de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_W) != 0) {
					// si pas de mur à l'ouest puis sud
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_W) == 0) &&
							((board[lig[idPlayer]][col[idPlayer]-1] & WALL_S) == 0) ) {
						return 0;
					}
				}
			}
			// si autre joueur au sud
			else if ((col[0] == col[1]) && (lig[idOther] == lig[idPlayer]+1)) {
				// si mur au sud de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_S) != 0) {
					// si pas de mur au sud puis à l'ouest
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_S) == 0) &&
							((board[lig[idPlayer]+1][col[idPlayer]] & WALL_W) == 0) ) {
						return 0;
					}
				}
			}
		}
		// saut oblique nord-ouest
		else if ((c == col[idPlayer]-1) && (l == lig[idPlayer]-1)) {
			// si autre joueur à l'ouest
			if ((lig[0] == lig[1]) && (col[idOther] == col[idPlayer]-1)) {
				// si mur à l'ouest de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_W) != 0) {
					// si pas de mur à l'ouest puis nord
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_W) == 0) &&
							((board[lig[idPlayer]][col[idPlayer]-1] & WALL_N) == 0) ) {
						return 0;
					}
				}
			}
			// si autre joueur au nord
			else if ((col[0] == col[1]) && (lig[idOther] == lig[idPlayer]-1)) {
				// si mur au nord de l'autre joueur
				if ( (board[lig[idOther]][col[idOther]] & WALL_N) != 0) {
					// si pas de mur au nord puis à l'ouest
					if ( ((board[lig[idPlayer]][col[idPlayer]] & WALL_N) == 0) &&
							((board[lig[idPlayer]-1][col[idPlayer]] & WALL_W) == 0) ) {
						return 0;
					}
				}
			}
		}

		// cas dplt normal
		// dplt possible lig[X]+/- 1 ou col[X]+/-1 si pas de mur
		if (l == lig[idPlayer]) {
			if (c == col[idPlayer]+1) {
				if ((board[l][c] & WALL_W) == 0) {
					return 0;
				}
			}
			else if (c == col[idPlayer]-1) {
				if ((board[l][c] & WALL_E) == 0) {
					return 0;
				}
			}
		}
		else if (c == col[idPlayer]) {
			if (l == lig[idPlayer]+1) {
				if ((board[l][c] & WALL_N) == 0) {
					return 0;
				}
			}
			else if (l == lig[idPlayer]-1) {
				if ((board[l][c] & WALL_S) == 0) {
					return 0;
				}
			}
		}

		return -1;
	}

	public int putWall(int idPlayer, int l, int c, int ori) {

		if (canPutWall(idPlayer, l, c, ori) == 0) {
			board[l][c] |= ori;
			if (ori == WALL_E) {
				board[l][c+1] |= WALL_W;
			}
			else if (ori == WALL_N) {
				board[l-1][c] |= WALL_S;
			}
			else if (ori == WALL_W) {
				board[l][c-1] |= WALL_E;
			}
			else if (ori == WALL_S) {
				board[l+1][c] |= WALL_N;
			}
			nbWalls[idPlayer] -= 1;
			return 0;
		}

		return -1;
	}

	/* test pour poser une barrière :
       possible :
          - s'il n'existe pas déjà une barrière à cet endroit.
          - si cela ne bouche pas totalement l'accès à l'autre bord.
       cas particuliers :
          - ligne 0 : impossible de mettre sauf au sud.
          - ligne 10 : impossible de mettre sauf au nord
     */
	public int canPutWall(int idPlayer, int l, int c, int ori) {

		int ret = -1;

		if (nbWalls[idPlayer] == 0) return -1;
		if ( (board[l][c] & ori) != 0) return -1;
		if ((l == 0) && (ori != WALL_S)) return -1;
		if ((l == 10) && (ori != WALL_N)) return -1;

		/* test si dplt possible jusqu'à l'autre bord.
		   cas n°1 : il existe un mur continu entre colonne 0 et colonne 8
		   au nord/sud du pion de l'autre joueur.
		   cas n°2 : ça enferme le pion de l'autre joueur

		   Il suffit de poser la barrière puis de faire un parcours en
		   largeur case par case pour voir si le joueur adverse peut
		   atteindre son arrivée.

		   Pour signaler qu'une case a été visitée, on met à 1 le bit
		   8 de gameState.
		 */
		int idOther = (idPlayer+1)%2;
		Queue<Integer> fifo = new ArrayDeque<>();
		// mise en place mur
		board[l][c] |= ori;
		int idCell = l*9+c;
		int ll,cc;
		fifo.offer(idCell);
		while(!fifo.isEmpty()) {
			idCell = fifo.poll();
			ll = idCell/9;
			cc = idCell%9;
			// test si sur ll = arrivée.
			if ((idOther == 0) && (ll == 10)){
				ret = 0;
				break;
			}
			else if ((idOther == 1) && (ll == 0)) {
				ret = 0;
				break;
			}
			// sinon voisins accessible non marqués dans fifo
			if ((board[ll][cc] & WALL_E) == 0) {
				if ((board[ll][cc+1] & 256) == 0) {
					fifo.offer(ll*9+cc+1);
				}
			}
			if ((board[ll][cc] & WALL_W) == 0) {
				if ((board[ll][cc-1] & 256) == 0) {
					fifo.offer(ll*9+cc-1);
				}
			}
			if ((board[ll][cc] & WALL_N) == 0) {
				if ((board[ll-1][cc] & 256) == 0) {
					fifo.offer((ll-1)*9+cc);
				}
			}
			if ((board[ll][cc] & WALL_S) == 0) {
				if ((board[ll+1][cc] & 256) == 0) {
					fifo.offer((ll+1)*9+cc);
				}
			}
			// marquer la case comme visitée
			board[ll][cc] |= 256;
		}

		// enlever les cases visitées
		for(int i=0;i<11;i++) {
			for(int j=0;j<9;j++) {
				board[i][j] ^= (board[i][j] & 256);
			}
		}
		// enlever le mur
		board[l][c] ^= (board[l][c] & ori);
		// si la fifo est vide, pas de solution donc ret vaut -1
		return ret;
	}

	private boolean isNeighbors() {

		if (lig[0] == lig[1]) {
			if (col[0] == (col[1]-1)) {
				// si pas mur est en lig[0],col[0] => ok
				if ( (board[lig[0]][col[0]] & WALL_E) == 0) {
					return true;
				}
			}
			else if (col[0] == (col[1]+1)) {
				// si pas mur est en lig[0],col[0] => ok
				if ( (board[lig[0]][col[0]] & WALL_W) == 0) {
					return true;
				}
			}
		}
		if (col[0] == col[1]) {
			if (lig[0] == (lig[1]-1)) {
				// si pas mur sud en lig[0],col[0] => ok
				if ( (board[lig[0]][col[0]] & WALL_S) == 0) {
					return true;
				}
			}
			else if (lig[0] == (lig[1]+1)) {
				// si pas mur nord en lig[0],col[0] => ok
				if ( (board[lig[0]][col[0]] & WALL_N) == 0) {
					return true;
				}
			}
		}
		return false;
	}

	public int[] possiblePawnMoves(int idPlayer) {
	/* 12 cas de dplt stockés dans le tableau dplt[] :
	   - dplt une seule case nord, sud, est, ouest
	   - dlt 2 cases nord, sub, est, ouest
	   - dplt oblique nord-est, nord-ouest, sud-est, sud-ouest

	   ca donne donc :
	     0,1     2,3      4,5     6,7      8,9     10,11   12,13   14,15    16,17     18,19     20,21     22,23
	   1 au N | 1 au S | 1 à E | 1 à O | 2 au N | 2 au S | 2 à E | 2 à O | 1 au NE | 1 au NO | 1 au SE | 1 au SO

	   le tableau fait 24 cases car chaque possibilité est donnée sur deux cases contenant
	   la ligne puis la colonne. Si les deux cases contiennent -1, cela veut dire que le déplacement
	   associé n'est pas possible compte tenu de l'état du plateau.
	 */
		int[] dplt = new int[24];
		for(int i=0;i<24;i++) dplt[i] = -1;
		// test une case nord
		if (canMovePawn(idPlayer,lig[idPlayer]-1,col[idPlayer]) == 0) {
			dplt[0] = lig[idPlayer]-1;
			dplt[1] = col[idPlayer];
		}
		// test une case sud
		if (canMovePawn(idPlayer,lig[idPlayer]+1,col[idPlayer]) == 0) {
			dplt[2] = lig[idPlayer]+1;
			dplt[3] = col[idPlayer];
		}
		// test une case est
		if (canMovePawn(idPlayer,lig[idPlayer],col[idPlayer]+1) == 0) {
			dplt[4] = lig[idPlayer];
			dplt[5] = col[idPlayer]+1;
		}
		// test une case ouest
		if (canMovePawn(idPlayer,lig[idPlayer],col[idPlayer]-1) == 0) {
			dplt[6] = lig[idPlayer];
			dplt[7] = col[idPlayer]-1;
		}
		// test deux cases nord
		if (canMovePawn(idPlayer,lig[idPlayer]-2,col[idPlayer]) == 0) {
			dplt[8] = lig[idPlayer]-2;
			dplt[9] = col[idPlayer];
		}
		// test deux cases sud
		if (canMovePawn(idPlayer,lig[idPlayer]+2,col[idPlayer]) == 0) {
			dplt[10] = lig[idPlayer]+2;
			dplt[11] = col[idPlayer];
		}
		// test deux cases est
		if (canMovePawn(idPlayer,lig[idPlayer],col[idPlayer]+2) == 0) {
			dplt[12] = lig[idPlayer];
			dplt[13] = col[idPlayer]+2;
		}
		// test deux cases ouest
		if (canMovePawn(idPlayer,lig[idPlayer],col[idPlayer]-2) == 0) {
			dplt[14] = lig[idPlayer];
			dplt[15] = col[idPlayer]-2;
		}
		// test une case nord-est
		if (canMovePawn(idPlayer,lig[idPlayer]-1,col[idPlayer]+1) == 0) {
			dplt[16] = lig[idPlayer]-1;
			dplt[17] = col[idPlayer]+1;
		}
		// test une case nord-ouest
		if (canMovePawn(idPlayer,lig[idPlayer]-1,col[idPlayer]-1) == 0) {
			dplt[18] = lig[idPlayer]-1;
			dplt[19] = col[idPlayer]-1;
		}
		// test une case sud-est
		if (canMovePawn(idPlayer,lig[idPlayer]+1,col[idPlayer]+1) == 0) {
			dplt[20] = lig[idPlayer]+1;
			dplt[21] = col[idPlayer]+1;
		}
		// test une case sud-ouest
		if (canMovePawn(idPlayer,lig[idPlayer]+1,col[idPlayer]-1) == 0) {
			dplt[22] = lig[idPlayer]+1;
			dplt[23] = col[idPlayer]-1;
		}
		return dplt;
	}



}
